// Copyright © 2023-24 Apple Inc.

#pragma once

// clang-format off
#include "mlx/types/half_types.h"
#include "mlx/types/complex.h"
#include "mlx/backend/cpu/unary_ops.h"
#include "mlx/backend/cpu/binary_ops.h"
// clang-format on

const char* get_prebuilt_preamble();
